from __future__ import annotations
from typing import TYPE_CHECKING, Type, TypeVar, cast
from types import ModuleType, SimpleNamespace
import inspect
import dataclasses

import click
import pydantic

from tgzr.cli.utils import TGZRCliGroup

from ..session import Session, get_default_session
from .exceptions import MissingRunNativeModule
from .app_info import ShellAppContext, ShellAppInfo, DefaultShellAppInfo

if TYPE_CHECKING:
    from ..session import Session


class ShellAppSettings(pydantic.BaseModel):
    pass


class NoSettings(ShellAppSettings):
    pass


SettingsType = TypeVar("SettingsType", bound=ShellAppSettings)


@dataclasses.dataclass
class ShellAppState[AppSettingsType: ShellAppSettings]:
    session: Session
    _app: _BaseShellApp[AppSettingsType]
    _current_workspace: str | None = None

    _app_context_cache: ShellAppContext | None = None
    _app_settings_cache: AppSettingsType | None = None

    data: SimpleNamespace = dataclasses.field(default_factory=SimpleNamespace)

    @property
    def app(self) -> _BaseShellApp[AppSettingsType]:
        return self._app

    @property
    def app_context(self) -> ShellAppContext:
        if self._app_context_cache is None:
            self._app_context_cache = self.app.create_app_context(self.session)
        return self._app_context_cache

    @property
    def settings_session_context(self) -> list[str]:
        layers = []
        if self.session.context.studio_name:
            layers.append(self.session.context.studio_name)
        if self.session.context.project_name:
            layers.append(self.session.context.project_name)
        if not layers:
            return []
        compated_layers = f"[{'/'.join(layers)}]"
        layers = [compated_layers]
        if self.session.context.entity_name:
            layers.append(self.session.context.entity_name)
        layers.append(self.session.context.user_name)
        return layers

    @property
    def app_settings_context(self) -> list[str]:
        return [
            *self.session.context.settings_base_context,
            *self.settings_session_context,
        ]

    def app_settings(self, reload: bool = False) -> AppSettingsType:
        if reload or self._app_settings_cache is None:
            self._app_settings_cache = self.app.get_settings(
                self.app_context, self.app_settings_context
            )
        return self._app_settings_cache


class _BaseShellApp[SettingsType: ShellAppSettings]:

    def __init__(
        self,
        app_name: str,
        run_module: ModuleType | None,
        app_id: str | None = None,
        app_groups: set[str] = set(),
        default_app_info: DefaultShellAppInfo | None = None,
        default_settings: SettingsType | None = None,
    ):
        """
        Create an App which will be accessible in the tgzr CLI.

        `app_name`:
            The name of the app, used to name the CLI command.

        `app_id`:
            The unic id associated with the app.
            The default value is the name of the module defining the
            app. This should be unic enough.

        `run_module`:
            The python module used to run the app.
            Must be a valid modules with appropriate safeguards.
            NOTE: using None will raise an exception showing you
            how to implement this module.

        `app_groups`:
            A set of groups where this app should be found.
            Groups with double-underscopre (like "_EXE_") are managed by tgzr.

        `app_info`:
            Provides the default info about this app.
            See _BaseShellApp.get_info()
        """
        frame: inspect.FrameInfo = inspect.stack()[1]
        if app_name is None:
            app_name = frame.function
            # print("---> AUTO APP NAME:", repr(app_name))

        if app_id is None:
            # FIXME: the module used here is wrong!
            module = inspect.getmodule(frame[0])
            app_id = (
                (module and module.__name__ or "???").replace(".", "_") + "." + app_name
            )
            # print("---> AUTO APP ID:", app_id)

        if run_module is None:
            raise MissingRunNativeModule(self)
        self.run_native_module = run_module

        self.app_name = app_name
        self.app_id = app_id
        self.app_groups = app_groups
        self._default_app_info = default_app_info or DefaultShellAppInfo()

        self._default_settings = default_settings or NoSettings()
        # FIXME: find a way to do this:
        # if not isinstance(self._default_settings, SettingsType):
        #     raise ValueError(
        #         f"The default settings must be an instance of {SettingsType}, not {type(default_settings)}"
        #     )

    def create_app_context(
        self,
        session: Session,
        host_suffix: str = "",
        context_name: str = "root",
    ):
        if host_suffix and not host_suffix.startswith("."):
            host_suffix = "." + host_suffix
        return ShellAppContext(
            session=session,
            host_name=f"tgzr.shell_apps.{self.app_id}{host_suffix}",
            context_name=context_name,
        )

    def create_app_state(self) -> ShellAppState[SettingsType]:
        """
        Return a state object that apps frontend can use to access
        various usefull things like the session, settings, etc...
        """
        session = get_default_session()
        if session is None:
            raise Exception("Oops, invalid session :/")
        app_state = ShellAppState[SettingsType](session=session, _app=self)
        return app_state

    def installed_versions(self, session: Session) -> set[str]:
        """
        Return the list of installed versions.
        Default is to return an empty set, which means this app
        does not support multiple versions
        """
        return set()

    def cli_run_cmd_installed(
        self, created_cmd: click.Command, root_group: TGZRCliGroup
    ):
        """
        Called when tgzr.shell.cli_plugin.app_cli has created and
        registered a cli command to execute this app.

        Subclasses can override this to alter the cmd or set it as default
        on the root group.

        Default does nothing.
        """
        pass

    def get_info(self, context: ShellAppContext) -> ShellAppInfo:
        """
        Subclasses can reimplement this to provide
        information used by GUIs to display (or hide) this app
        in the given context.

        The default behavior is to return a copy of the `app_info`
        provided in the constructor, and hide the app if
        context.context_name is not part of the app.app_groups set.
        """
        # print("???", self.app_name, self.app_groups, context.context_name)
        app_info = ShellAppInfo(
            app=self,
            title=self._default_app_info.title or self.app_name.title(),
            icon=self._default_app_info.icon,
            color=self._default_app_info.color,
            hidden=context.context_name not in self.app_groups,
            installed_versions=self.installed_versions(context.session),
        )
        return app_info

    @property
    def settings_key(self) -> str:
        return f"shell_apps.{self.app_id}"

    async def get_settings(
        self, context: ShellAppContext, settings_context: list[str] | None = None
    ) -> SettingsType:
        settings_context = settings_context or []

        model_type = type(self._default_settings)
        settings = await context.session.settings.get_context(
            settings_context,
            cast(Type[pydantic.BaseModel], model_type),
            self.settings_key,
        )
        return cast(SettingsType, settings)

    async def store_settings(
        self,
        settings: SettingsType,
        context: ShellAppContext,
        context_name: str | None = None,
        exclude_defaults: bool = True,
    ) -> None:
        if not isinstance(settings, pydantic.BaseModel):
            raise Exception()
        # If not provided, we right in the current user's context:
        context_name = context_name or context.session.context.user_name

        await context.session.settings.update_context(
            context_name=context_name,
            model=settings,
            path=self.settings_key,
            exclude_defaults=exclude_defaults,
        )

    def run_app(self, session: Session):
        raise NotImplementedError()



from typing import Literal

import importlib.metadata

import click
import rich
import rich.table

from tgzr.cli.utils import TGZRCliGroup

from ..session import Session
from ..studio import Studio
from ..project import Project

from .utils import pass_session


@click.group(
    cls=TGZRCliGroup,
    help="Manage Projects in the Studio",
)
def project():
    pass


@project.command()
@click.argument("name")
@click.option("--allow-existing", is_flag=True, default=False)
@click.option(
    "-r",
    "--required-packages",
    multiple=True,
    help='Extra packages to install. Can be set multiple times: -r "my_studio==1.2.3" -r "krita blender kitsu"',
)
@click.option(
    "--default-index",
    help="The URL of the default package index (by default: <https://pypi.org/simple>).",
)
@click.option(
    "-f",
    "--find-links",
    help="path a folder containing packages to install. Usefull in no-internet situations.",
)
@click.option(
    "--allow-prerelease",
    is_flag=True,
    help="Allow installing tgzr using pre-release packages. Default is False.",
)
@pass_session
def create(
    session: Session,
    name,
    allow_existing: bool = False,
    required_packages: list[str] = [],
    default_index: str | None = None,
    find_links: str | None = None,
    allow_prerelease: bool = False,
):
    """
    Create a new project with name "NAME" in the current Studio.
    """
    try:
        studio: Studio | None = session.get_selected_studio()
    except FileNotFoundError:
        raise click.UsageError(
            f'Studio "{session.selected_studio_name}" does not exists.'
        )
    if studio is None:
        raise click.UsageError("Please select a studio first.")

    project = studio.get_project(name)
    if project.exists():
        if not allow_existing:
            raise click.UsageError(
                f'A Project "{name}" already exists in studio "{studio.path}". Use --allow-existing to update it.'
            )

    studio.create_project(
        project_name=name,
        required_packages=required_packages,
        index=default_index,
        find_links=find_links,
        allow_prerelease=allow_prerelease,
    )
    click.echo(f"Project {name!r} created.")


@project.command()
@pass_session
def show(session: Session):
    """Show the project config."""
    try:
        studio: Studio | None = session.get_selected_studio()
    except FileNotFoundError:
        raise click.UsageError(
            f'Studio "{session.selected_studio_name}" does not exists.'
        )
    if studio is None:
        raise click.UsageError(
            "No Studio selected, use `tgzr --studio <name> ...` to specify one, or `tgzr studio select` to set the default one."
        )
    project: Project | None = session.get_selected_project()
    if project is None:
        raise click.UsageError(
            "No Project selected, use `tgzr --project <name>...` to specify one."
        )
    if not project.exists():
        raise click.UsageError(f'Project "{project.name}" does not exists.')

    click.echo(f'Configuration for Project "{project.name}" ({project.path}):')
    rich.print(project.config)


@project.command()
@pass_session
def packages(session: Session):
    """Show the project packages."""
    try:
        studio: Studio | None = session.get_selected_studio()
    except FileNotFoundError:
        raise click.UsageError(
            f'Studio "{session.selected_studio_name}" does not exists.'
        )
    if studio is None:
        raise click.UsageError(
            "No Studio selected, use `tgzr --studio <name> ...` to specify one, or `tgzr studio select` to set the default one."
        )
    project: Project | None = session.get_selected_project()
    if project is None:
        raise click.UsageError(
            "No Project selected, use `tgzr --project <name>...` to specify one."
        )
    if not project.exists():
        raise click.UsageError(f'Project "{project.name}" does not exists.')

    dists = project.get_packages()
    table = rich.table.Table(
        "Name",
        "Version",
        "Editable Path",
        title=f'Packages in Project "{project.name}" ({project.path})',
        min_width=80,
    )
    for dist in dists:
        editable_path = ""
        if dist.origin:
            editable_path = dist.origin.dir_info.editable
        table.add_row(dist.name, dist.version, editable_path)
    rich.print(table)


@project.command()
@click.option(
    "-f",
    "--group-filter",
    help="Show only this type of plugins (defaults is to show all).",
)
@pass_session
def plugins(session: Session, group_filter: str | None):
    """Show the project plugins."""
    try:
        studio: Studio | None = session.get_selected_studio()
    except FileNotFoundError:
        raise click.UsageError(
            f'Studio "{session.selected_studio_name}" does not exists.'
        )
    if studio is None:
        raise click.UsageError(
            "No Studio selected, use `tgzr --studio <name> ...` to specify one, or `tgzr studio select` to set the default one."
        )
    project: Project | None = session.get_selected_project()
    if project is None:
        raise click.UsageError(
            "No Project selected, use `tgzr --project <name>...` to specify one."
        )
    if not project.exists():
        raise click.UsageError(f'Project "{project.name}" does not exists.')

    plugins = project.get_plugins(group_filter)
    table = rich.table.Table(
        "Group",
        "Name",
        "Value",
        "Package",
        title=f'Plugins in Project "{project.name}" ({project.path})',
        min_width=80,
    )
    for plugin, distribution in plugins:
        package = f"{distribution.name} {distribution.version}"
        editable = distribution.origin and distribution.origin.dir_info.editable
        if editable:
            package = f"[orange1]{package}[/orange1]"
        table.add_row(
            plugin.group,
            plugin.name,
            plugin.value,
            package,
        )
    rich.print(table)


@project.command(
    "run",
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@click.argument(
    "cmd_name",
)
@click.argument("cmd_args", nargs=-1, type=click.UNPROCESSED)
@pass_session
def project_run(session: Session, cmd_name: str, cmd_args: list[str]):
    """
    Run a Project cmd.
    """
    try:
        studio: Studio | None = session.get_selected_studio()
    except FileNotFoundError:
        raise click.UsageError(
            f'Studio "{session.selected_studio_name}" does not exists.'
        )
    if studio is None:
        raise click.UsageError("Please select a studio first.")
    project: Project | None = session.get_selected_project()
    if project is None:
        raise click.UsageError("Please select a project first.")
    studio.run_cmd(cmd_name, cmd_args, project.name)


@project.command
@click.option(
    "-s",
    "--shell",
    default="auto",
    help='The shell to use, one of ["xonsh", "cmd", "powershell", "auto"] (defaults to "auto").',
)
@pass_session
def shell(
    session: Session,
    shell: Literal["xonsh", "cmd", "powershell", "auto"] = "auto",
):
    """
    Open a shell configured for the studio.
    """
    try:
        studio: Studio | None = session.get_selected_studio()
    except FileNotFoundError:
        raise click.UsageError(
            f'Studio "{session.selected_studio_name}" does not exists.'
        )
    if studio is None:
        raise click.UsageError("Please select a studio or set a default one.")
    project: Project | None = session.get_selected_project()
    if project is None:
        raise click.UsageError("Please select a project first.")
    if not project.exists():
        raise click.UsageError(f"Project {project.name} does not exist.")
    studio.shell(shell_type=shell, project_name=project.name)



from nicegui import ui


@ui.page("/", title="TGZR - Manager Panel", dark=True)
async def main():
    ui.label("Yeah YOLO! 😛").classes("text-h5")

    def do_it(e):
        import os

        l = os.listdir("/")
        label.text = ", ".join(l)

    label = ui.label().classes("tracking-widest")
    with ui.column():
        ui.label("go").classes("bg-blue-800 m-5 p-5 rounded-xl")
        ui.label("go")
        ui.label("go")
        ui.label("go")
        ui.label("go adsf a asdfs")
        ui.label("go")
        ui.label("go")
        ui.button("Go", on_click=do_it)
